import numpy as np
import pandas as pd
import tensorflow as tf 
import re
import pandas as pd
from sklearn.feature_extraction.text import TfidfVectorizer
from sklearn.neighbors import NearestNeighbors
from sklearn.metrics.pairwise import cosine_similarity
import keyboard

dataset = pd.read_csv('data/que_ans.csv')

dataset

dataset = dataset[['Question', 'Answer']]

dataset

def preprocess_text(text):
    # Remove special characters and convert to lowercase
    text = re.sub(r"[^a-zA-Z0-9]", " ", text).lower()
    return text

# Apply preprocessing to the answer column
dataset['Answer'] = dataset['Answer'].apply(preprocess_text)

data = dataset

data

from keras.utils import pad_sequences
# Text tokenization
tokenizer = tf.keras.preprocessing.text.Tokenizer()
tokenizer.fit_on_texts(data['Answer'])
encoded_answer = tokenizer.texts_to_sequences(data['Answer'])

# Padding sequences
max_seq_length = max(len(seq) for seq in encoded_answer)
padded_answer = pad_sequences(encoded_answer, maxlen=max_seq_length, padding='post')

padded_answer

# Vectorize the existing answers
vectorizer = TfidfVectorizer()
answer_vectors = vectorizer.fit_transform(data['Answer'])

# Generate 20 random questions
random_indices = np.random.randint(0, len(data), size=2)
random_questions = data['Question'][random_indices]
random_answers = data['Answer'][random_indices]

correct_count = 0
incorrect_questions = []

for question, answer in zip(random_questions, random_answers):
    print(answer)
    print('Question:', question)
    user_answer = ""
    
    # Close answer palette when 'q' is pressed
    while True:

        if keyboard.is_pressed('q'):
            break
        elif keyboard.is_pressed('enter'):
            user_answer = input("Enter your answer: ")
            break

    if not user_answer:
        break
    
    random_answer_vector = vectorizer.transform([answer])
    new_answer_vector = vectorizer.transform([user_answer])
    
    similarity_scores = cosine_similarity(random_answer_vector, new_answer_vector)
    
    if similarity_scores >= 0.4:
        print('Correct (Similarity Score:', similarity_scores, ')')
        correct_count += 1
    else:
        print('Incorrect (Similarity Score:', similarity_scores, ')')
        incorrect_questions.append(question)

print('Marks:', correct_count*20,"%\n\n")
#print('Incorrect Questions:\n\n', incorrect_questions)

incorrect_questions

topics_df = pd.read_csv('data/topics.csv')

topics = topics_df['topic']

topics

# Vectorize the user answer
topic_vector = vectorizer.fit_transform(topics_df['topic'])

# Use nearest neighbors to find the most similar answers
neigh = NearestNeighbors(n_neighbors=3, metric='cosine')
neigh.fit(topic_vector)

ranked_topics = []
for question in incorrect_questions:
    question_vector = vectorizer.transform([question])  # Pass the question as a list
    distances, indices = neigh.kneighbors(question_vector.reshape(1, -1))  # Reshape the question_vector

    # Get the ranked answers
    ranked_topic = topics_df['topic'].iloc[indices[0]]
    ranked_topics.append(ranked_topic)

# Print the similar topics for the student to follow
print('You may consider exploring the following topics:\n')
for topic in ranked_topics:
    print(topic)




