from flask import Flask, render_template, url_for, redirect, request, jsonify
from flask_sqlalchemy import SQLAlchemy
from flask_login import UserMixin, login_user, LoginManager, login_required, logout_user, current_user
from flask_wtf import FlaskForm
from wtforms import StringField, PasswordField, SubmitField
from wtforms.validators import InputRequired, Length, ValidationError
from flask_bcrypt import Bcrypt
from similarity import perform_ocr_and_similarity
from flask import Flask, render_template,session
import numpy as np
import pandas as pd
from sklearn.feature_extraction.text import TfidfVectorizer
from sklearn.metrics.pairwise import cosine_similarity
from sklearn.neighbors import NearestNeighbors
import os
import cv2
import keras_ocr

app = Flask(__name__)
bcrypt = Bcrypt(app)
app.config['SQLALCHEMY_DATABASE_URI'] = 'sqlite:///database.db'
app.config['SECRET_KEY'] = 'thisisasecretkey'
app.config['MAX_CONTENT_LENGTH'] = 16 * 1024 * 1024  # Set max upload size to 16 MB

db = SQLAlchemy(app)


login_manager = LoginManager()
login_manager.init_app(app)
login_manager.login_view = 'login'


@login_manager.user_loader
def load_user(user_id):
    return User.query.get(int(user_id))


class User(db.Model, UserMixin):
    id = db.Column(db.Integer, primary_key=True)
    username = db.Column(db.String(20), nullable=False, unique=True)
    password = db.Column(db.String(80), nullable=False)


class RegisterForm(FlaskForm):
    username = StringField(validators=[
                           InputRequired(), Length(min=4, max=20)], render_kw={"placeholder": "Username"})

    password = PasswordField(validators=[
                             InputRequired(), Length(min=8, max=20)], render_kw={"placeholder": "Password"})

    submit = SubmitField('Register')

    def validate_username(self, username):
        existing_user_username = User.query.filter_by(
            username=username.data).first()
        if existing_user_username:
            raise ValidationError(
                'That username already exists. Please choose a different one.')


class LoginForm(FlaskForm):
    username = StringField(validators=[
                           InputRequired(), Length(min=4, max=20)], render_kw={"placeholder": "Username"})

    password = PasswordField(validators=[
                             InputRequired(), Length(min=8, max=20)], render_kw={"placeholder": "Password"})

    submit = SubmitField('Login')


@app.route('/')
def home():
    return render_template('home.html')


@app.route('/login', methods=['GET', 'POST'])
def login():
    form = LoginForm()
    if form.validate_on_submit():
        user = User.query.filter_by(username=form.username.data).first()
        if user:
            if bcrypt.check_password_hash(user.password, form.password.data):
                login_user(user)
                return redirect(url_for('dashboard'))
    return render_template('login.html', form=form)


USER_ROLES = {
    "student1": "student",
    "teacher1": "teacher",
    # Add more username-role mappings as needed
}

@app.route('/dashboard', methods=['GET', 'POST'])
@login_required
def dashboard():
    username = current_user.username
    user_role = USER_ROLES.get(username, "default_role")  # Assign a default role if username not found in the mapping
    print(username)
    return render_template('dashboard.html', user_role=user_role)

# Import necessary libraries and functions (including your ER diagram generation logic)
from wtforms import StringField, SubmitField
from wtforms.validators import DataRequired

class ERDiagramForm(FlaskForm):
    user_input = StringField('User Input', validators=[DataRequired()])
    submit = SubmitField('Generate ER Diagram')

import subprocess
import json

@app.route('/generate_er_diagram', methods=['GET', 'POST'])
@login_required
def generate_er_diagram():
    form = ERDiagramForm(request.form)
    sql_queries = ""

    if form.validate_on_submit():
        user_input = form.user_input.data
        user_input = user_input.replace('"', '')  # Process the input as needed

        # Call the diagram.py script with the processed user input
        try:
            result = subprocess.run(["python", "diagram.py", user_input], capture_output=True, text=True, check=True)
            sql_queries = result.stdout.strip()  # Get the SQL queries from the subprocess output
        except Exception as e:
            sql_queries = f"Error: {e}"

        # Load SQL queries from the JSON file
        with open('data/sql_queries.json', 'r') as json_file:
            saved_queries = json.load(json_file)

        return render_template('er_diagram.html', form=form, er_diagram=True, sql_queries=saved_queries)

    return render_template('er_diagram.html', form=form, er_diagram=False, sql_queries=sql_queries)

import cv2
from flask import send_from_directory

# Get the absolute path to the temp folder
temp_folder_path = os.path.abspath('temp/')

@app.route('/images/<filename>')
def get_image(filename):
    return send_from_directory(temp_folder_path, filename)

# Route to handle image similarity functionality
@app.route('/similarity', methods=['GET', 'POST'])
@login_required
def similarity():
    temp_folder = 'temp'
    os.makedirs(temp_folder, exist_ok=True)  
    
    if request.method == 'POST':
        # Handle uploaded images
        image1 = request.files['image1']
        image2 = request.files['image2']
        
        # Save the uploaded images to temporary files
        image1_path = os.path.join(temp_folder, 'image1.jpg') 
        image2_path = os.path.join(temp_folder, 'image2.jpg') 
        image1.save(image1_path)
        image2.save(image2_path)
        
        # Invert the images
        #inverted_image1 = cv2.bitwise_not(cv2.imread(image1_path))
        #inverted_image2 = cv2.bitwise_not(cv2.imread(image2_path))

        img = cv2.imread(image1_path)
        img2 = cv2.imread(image2_path)
        
        # Load the images
        inverted_image1 = cv2.bitwise_not(img)
        inverted_image2 = cv2.bitwise_not(img2)

        # Save the inverted images with correct names
        inverted_image1_path = os.path.join(temp_folder, 'inverted_image1.jpg')
        inverted_image2_path = os.path.join(temp_folder, 'inverted_image2.jpg')
        cv2.imwrite(inverted_image1_path, inverted_image1)
        cv2.imwrite(inverted_image2_path, inverted_image2)
    
        # Perform OCR and similarity calculations (assuming these functions are defined)
        word_list1, corrected_word_list1, word_list2, corrected_word_list2, pred, corr = perform_ocr_and_similarity(inverted_image1_path, inverted_image2_path)
        
        # Call the similarity.py script with the processed image paths
        try:
            result = subprocess.run(["python", "similarity.py", inverted_image1_path, inverted_image2_path], capture_output=True, text=True, check=True)
            similarity_result = result.stdout.strip()
        except Exception as e:
            similarity_result = f"Error: {e}"
        
        similarity_result = "Similarity Result"
        # Detailed results dictionary
        detailed_results = {
            "word_list1": word_list1,
            "corrected_word_list1": corrected_word_list1,
            "word_list2": word_list2,
            "corrected_word_list2": corrected_word_list2,
            "pred": pred,
            "corr": corr
        }
        
        return render_template('similarity.html', similarity_result=similarity_result, detailed_results=detailed_results)

    return render_template('similarity.html')

import numpy as np
from spellchecker import SpellChecker
from nltk.stem import SnowballStemmer

pipeline = keras_ocr.pipeline.Pipeline()
spell = SpellChecker()
stemmer = SnowballStemmer("english")

# Calculate marks, marks1, marks2, marks3, and final_marks here...
def calculate_marks(corrected_words):
    with open('data/entity_list.json', 'r') as json_file:
        entity_list = json.load(json_file)
    with open('data/attributes.json', 'r') as json_file:
        attributes_data = json.load(json_file)
    with open('data/relations.json', 'r') as json_file:
        relations_data = json.load(json_file)
    
    entity_matches = sum(word.lower() in entity_list for word in corrected_words)

    # Filter out None values from corrected_words list
    corrected_words = [word for word in corrected_words if word is not None]
    
    attribute_words = [attribute.split('_')[-1].lower() for attributes in attributes_data.values() for attribute in attributes]
    attribute_matches = sum(word.lower() in attribute_words for word in corrected_words)
    
    relation_words = [stemmer.stem(word.split()[0]) for word in relations_data.keys()]
    relation_matches = sum(stemmer.stem(word) in relation_words for word in corrected_words)
    
    total_entities = len(entity_list)
    total_attributes = len(attribute_words)
    total_relations = len(relation_words)
    
    marks1 = entity_matches / total_entities
    marks2 = attribute_matches / total_attributes
    marks3 = relation_matches / total_relations
    
    final_marks = (marks1 + marks2 + marks3) / 3 * 10
    return marks1, marks2, marks3, final_marks

import traceback

@app.route('/extract', methods=['GET', 'POST'])
@login_required
def extract():
    if request.method == 'POST':
        try:
            image_file = request.files['image']
            
            if image_file.filename == '':
                return "No selected file"
            
            # Process the file
            img = cv2.imdecode(np.fromstring(image_file.read(), np.uint8), cv2.IMREAD_COLOR)
            inverted_image = cv2.bitwise_not(img)
            cv2.imwrite("temp/extract/inverted_image1.jpg", inverted_image)

            results = pipeline.recognize(["temp/extract/inverted_image1.jpg"])
            word_list = [word_info[0] for word_info in results[0]]
            corrected_word_list = [spell.correction(word) for word in word_list]

            # Load JSON data
            with open('data/entity_list.json', 'r') as json_file:
                json_word_list = json.load(json_file)
                json_word_list = [word.lower() for word in json_word_list]

             # Filter out None values from corrected_word_list
            corrected_word_list = [word.lower() for word in corrected_word_list if word is not None]

            pred_list = [word.lower() for word in corrected_word_list]
            matches = [word for word in json_word_list if word in pred_list]

            marks1, marks2, marks3, final_marks = calculate_marks(corrected_word_list)

            result = {
                'words': word_list,
                'corrected_words': corrected_word_list,
                'matched_entities': matches,
                'marks1': marks1,  # Calculate marks1
                'marks2': marks2,  # Calculate marks2
                'marks3': marks3,  # Calculate marks3
                'final_marks': final_marks  # Calculate final_marks
            }

            return render_template('extract.html', result=result)

        except Exception as e:
            # Log the exception for debugging purposes
            traceback.print_exc()  # Print the traceback to console
            return f"Error: {e}"

    return render_template('extract.html')

import pandas as pd
from sklearn.feature_extraction.text import TfidfVectorizer
from sklearn.metrics.pairwise import cosine_similarity
from sklearn.neighbors import NearestNeighbors
import re

def preprocess_text(text):
    return re.sub(r"[^a-zA-Z0-9]", " ", text).lower()

# Refactored initialization code from ai_quiz.py
def initialize_quiz_data():
    dataset = pd.read_csv('data/que_ans.csv')
    dataset = dataset[['Question', 'Answer']]
    dataset['Answer'] = dataset['Answer'].apply(preprocess_text)
    
    vectorizer = TfidfVectorizer()
    answer_vectors = vectorizer.fit_transform(dataset['Answer'])
    
    topics_df = pd.read_csv('data/topics.csv')
    topic_vector = vectorizer.fit_transform(topics_df['topic'])
    
    return dataset, vectorizer, answer_vectors, topic_vector, topics_df

data, vectorizer, answer_vectors, topic_vector, topics_df = initialize_quiz_data()

def evaluate_answers(user_answers, random_questions):
    correct_count = 0
    incorrect_questions = []
    similarity_scores_list = []  # List to hold similarity scores for each answer
    
    for question, user_answer in zip(random_questions, user_answers):
        correct_answer = data[data['Question'] == question]['Answer'].iloc[0]
        correct_answer_vector = vectorizer.transform([correct_answer])
        user_answer_vector = vectorizer.transform([user_answer])
        
        similarity_scores = cosine_similarity(correct_answer_vector, user_answer_vector)
        similarity_scores_list.append(similarity_scores[0][0])  # Add the similarity score to the list
        
        if similarity_scores >= 0.4:
            correct_count += 1
        else:
            incorrect_questions.append(question)
            
    # Recommend topics based on incorrect answers
    neigh = NearestNeighbors(n_neighbors=3, metric='cosine')
    neigh.fit(topic_vector)
    ranked_topics = []
    
    for question in incorrect_questions:
        question_vector = vectorizer.transform([question])
        _, indices = neigh.kneighbors(question_vector.reshape(1, -1))
        ranked_topic = topics_df['topic'].iloc[indices[0]]
        ranked_topics.extend(ranked_topic)
    
    return correct_count, similarity_scores_list, ranked_topics


@app.route('/quiz', methods=['GET', 'POST'])
@login_required
def quiz():
    if request.method == 'POST':
        if 'user_answer' in request.form:
            user_answers = request.form.getlist('user_answer')
            random_questions = session['questions']
            correct_count, similarity_scores_list, ranked_topics = evaluate_answers(user_answers, random_questions)
            return render_template('quiz.html', step='results', user_answers=user_answers, 
                                   correct_count=correct_count, questions=random_questions, 
                                   similarity_scores_list=similarity_scores_list, recommended_topics=ranked_topics,
                                   zip=zip)
    else: # This is the GET request, which means the student is starting the quiz
        num_questions = session.get('num_questions', 10)  # default to 10 if not set
        random_indices = np.random.randint(0, len(data), size=int(num_questions))
        random_questions = data['Question'].iloc[random_indices].tolist()
        session['questions'] = random_questions
        return render_template('quiz.html', step='answer', questions=random_questions)

    # This line will probably never be reached now, but let's keep it just in case
    return render_template('quiz.html', step='select_num_questions')


@app.route('/quiz_count', methods=['GET', 'POST'])
def quiz_count():
    if request.method == 'POST':
        num_questions = request.form['num_questions']
        topics = request.form.getlist('topics')  # Assuming topics is a multi-select field
        session['num_questions'] = num_questions
        session['topics'] = topics
        session['message'] = f"Topics {', '.join(topics)} were selected and the number of questions was set as {num_questions}."
        return redirect(url_for('quiz_count'))
    
    # Clear the message after it has been set
    message = session.pop('message', None)
    return render_template('quiz_count.html', message=message)


@app.route('/logout', methods=['GET', 'POST'])
@login_required
def logout():
    logout_user()
    return redirect(url_for('login'))


@ app.route('/register', methods=['GET', 'POST'])
def register():
    form = RegisterForm()

    if form.validate_on_submit():
        hashed_password = bcrypt.generate_password_hash(form.password.data)
        new_user = User(username=form.username.data, password=hashed_password)
        db.session.add(new_user)
        db.session.commit()
        return redirect(url_for('login'))

    return render_template('register.html', form=form)


if __name__ == "__main__":
    with app.app_context():
        db.create_all()  # Create the tables inside the application context
    app.run(debug=True)
